/*
 * Decodage Pulsadis EDF trame CPL 175 Hz 40 Bits
*/

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <Arduino.h>

#include "config.h"


/*
 * sampling rate is 700 Hz CK = 7.3728 Mhz
 * count = 7372800 / 700 = 10532
 */
/* timer 1 precaler setting : CK */
/* #define  F_CPU  7372800L */
#define  TC1_COUNT (65536L - F_CPU / SAMPLING_RATE)
#define  TCCR1A_SET   0
#define  TCCR1B_SET   (_BV(CS10))
#define  TIMSK_SET (_BV(TOIE1))
#define  ADMUX_SET (_BV(ADLAR))  /* select channel 0 */
 /* CK/64 */
#define ADCSRA_SET (_BV(ADEN) | _BV(ADSC) | _BV(ADIF) | _BV(ADPS2) | _BV(ADPS1))
/* timeout 2 seconds */
#define WDTCR_SET (_BV(WDE) | _BV(WDP2) | _BV(WDP1) | _BV(WDP0))

/*
 * Local prototypes
 */
//int detect_175(void);
uint8_t parse_msg(void);
/*
 * Global variables
 */
//volatile uint8_t silent;
int16_t s_min;
int16_t s_max;
volatile int8_t time_updated;
volatile int det_175;
volatile int m_det_175;
volatile uint16_t ref_clock;
volatile uint16_t ref_time;
volatile uint16_t delay_60s;
volatile uint8_t timer_100ms;
volatile uint16_t count_100ms;
volatile uint8_t timer_second;
volatile uint8_t seconds;
volatile uint16_t curtime; /* current time = hour * 60 + min */
volatile uint16_t tmp_result;

volatile uint8_t  rdata[5];
volatile uint8_t  cpdata[5];
volatile uint8_t  M_data[5];

/* 0 si on a recu une sequence apres reset, 0x80 sinon */
volatile uint8_t start_flag;  
volatile uint8_t first_msg_done;

uint8_t code_acde;

volatile uint16_t seq_result; /* bits of the last sequence */
uint8_t  msg_result[2];
uint8_t  msg_s_buf[2];
int flag_time = false;

  uint8_t count = 0;
  uint8_t nbset = 0;
  uint8_t msg_bit_num = 0;
  uint16_t start_time = 0;
  uint16_t next_step = 0;
  uint8_t ret ;
  uint8_t seuil = 511;
  int16_t ret1 ;
  uint8_t state = D_REPOS;
  
/*
 * timer 0 interrupt 700 Hz
 */

ISR(TIMER1_OVF_vect)
{
  noInterrupts();
     /* reload counter */
    TCNT1 = TC1_COUNT;
   /* Check if adc has complete.  */
   if (  bit_is_set(ADCSRA, ADIF) ){
      det_175 = detect_175();
      decode_175();
   }
   /* start new sampling cycle */
   ADCSRA = ADCSRA_SET; /* outp( ADCSRA_SET, ADCSRA) */
  ref_clock++;  /* 700 Hz */
  if ( (ref_clock & 1) == 0 ){
    ref_time++; /* 350 Hz */
    if ( ++timer_100ms >= SAMPLING_RATE / 10 / 2){ /* 700 / 70  = 10 hz */
      timer_100ms = 0;
      count_100ms++;
      if ( ++timer_second >= 10 ) {
        wdt_reset();
        timer_second = 0;
        /* time_update(); */
        if ( ++seconds >= 60 ) {
          seconds = 0;
          curtime++;
          if ( curtime >= 24 * 60 ) {
            curtime = 0;
          }
          time_updated++;
        }
      }
    }
  }
  interrupts();
}

int16_t detect_175(void)
{
  static int idx ; /* index for decoding algorythm */
  static int16_t ss0;
  static int16_t ss1;
  int16_t sample;

  if ( idx == 0 ) {
    ss0 = ss1 = 0;
  }
  /* Read the adc channel.  */
  /* sample = ADCL; * do not care with ADLAR = 1 - must be read first */
  sample = analogRead(A0); /* read adc register  */
  s_min = min(sample, s_min);
  s_max = max(sample, s_max);

  sample -= seuil;
  if ( (idx & 2) == 2 ) {
    /* bit 2 et 3 */
    sample = - sample;
  }
  if ( (idx & 1) == 0 ) {
    /* bit 0 et 2 */
    ss0 += sample;
  } 
  else {
    /* bit 1 et 3 */
    ss1 += sample;
  }
  if ( ++idx < NSAMPLES ) {
    return -1;
  }
  if ( ss0 < 0 ){
    ss0 = - ss0;
  }
  if ( ss1 < 0 ){
    ss1 = - ss1;
  }
  ss0 += ss1;
  idx = 0;

  return ss0;
}   
   
void setup() {
  Serial.begin(115200);
  noInterrupts();
     /* set timer 1 */
   TCCR1A = TCCR1A_SET;
   TCCR1B = TCCR1B_SET;
   TCNT1 = TC1_COUNT;
   TIMSK1 = TIMSK_SET;    /* timer_enable_int(TIMSK_SET) */
   /* set adc */
   ADMUX = ADMUX_SET;
   ADCSRA = ADCSRA_SET; /* start new sampling cycle */

   /* set watchdog */
   wdt_enable(WDTO_2S );
   wdt_reset();
  interrupts();
  initTempo();
}

void welcome(char end)
{
  String message;
  message = "";
  //message += F_CPU; /* welcome */
  message += "*Pulsadis*";
  //message += TC1_COUNT; /* welcome */
  if ( end) {
    message += end;
  }
    Serial.println(message);
}

void initTempo()
{
  // Serial.println("Init tempo");
  uint8_t i;
  det_175 = -1;
  welcome('\n');
  s_min = 1023;
 
  // timer_inp = count_100ms + 1 ;
  delay_60s =  600; // (1 << 9); /* delay 60 secondes */
  start_flag = 0x80;
  cpdata[0] = 0xA1;
  cpdata[1] = 0x24;
  read_last_seq();
  set_rjn();
  
}
/*
 * store the received bit in data array
 */
void tempo_bit_store(uint8_t msg_bit_num)
{
  rdata[msg_bit_num / 8] |= ( 0x80 >> (msg_bit_num % 8));
  return;
}

void tempo_clear_rdata(void)
{
  int i;
  for ( i = 0 ; i < NTBITS / 8 ; i++){
    cpdata[i] = rdata[i];
    rdata[i] = 0;
  }
}

String send_hour(uint16_t curtim, char *end)
{
  String message;
  char msg[2]; 
  uint8_t h;
  uint8_t m;   
  h = curtim / 60;
  m =  curtim % 60;
  message = "";
  /* rs_send(printd( h, 2, ' ') ); */
  sprintf(msg,"%2d",h);  
  message += String(msg);
  /* rs_send_char(':'); */
  message += ":";
  /* rs_send(printd( m, 2, '0') ); */
  sprintf(msg,"%2d",m);
  message += String(msg);
  if ( end ) {
    /* rs_send(end); */
    message += end;
  } 
  else {
    /* rs_send_char('\n'); */
    message += "\t";
  }
  return message;
}

/*
 * reception message de la liaison serie
 */
uint8_t  parse_msg(void)
{
      /* delay sending of first message */
  if ( first_msg_done == 0 && count_100ms == delay_60s ){
    //first_msg_done++;
    count_100ms = 0;
    read_last_seq();
    set_rjn();
  } 
  // Serial.println("parse_msg");
  char *pmsg = 0;
  int c_recu = 0;
  if ( ( c_recu = Serial.read()) == -1 ) {
    return 0;
  }
  *pmsg = char(c_recu);
  if( *pmsg == 'L' ) {  /* read last sequence */
    read_last_seq();
    set_rjn();
  } 
  return 1;
}


void read_last_seq(void)
{
  uint8_t i;
  String message;
  message = "";
  message += send_hour(curtime,0);

  message += "-" + Hexa2(cpdata[0]);
  message += "-" + Hexa2(cpdata[1]);
  message += "-" + Hexa2(cpdata[2]);
  message += "-" + Hexa2(cpdata[3]);
  message += "-" + Hexa2(cpdata[4]);
  /* rs_send_char('\n'); */
  message += "-ACDE:";
  message += Hexa2(code_acde); //String(code_acde,HEX);
  message += "-Min:";
  message += String(s_min);
  message += "-Max:";
  message += String(s_max) + "---";
  Serial.print(message);
}

// Conversion Hexa en ASCII 
String Hexa2(long hexa) // 
{
  String z;
  if (String(hexa,HEX).length() < 2) {
  z = "0" + String(hexa,HEX);
  z.toUpperCase();
  }
  else 
  {
  z = String(hexa,HEX);
  z.toUpperCase();
  }
  return z;
}

//
// Decodage Couleur Tempo alerte du Lendemain & Jour
// Alerte : Bleu bit 18, Blanc bit 21, Rouge bit 24
// Jour : Bleu bit 33, Blanc bit 36, Rouge bit 39
// Decodage Tarif Heure creuse 
// Heure Creuse Rouge bit 23, Heure Pleine Rouge bit 26
// Tarif Nuit tempo bit 32, Tarif Jour tempo bit 35
//
void set_tempo(void)
{
  //Serial.println("Set tempo");
  uint32_t m_tempo;
  String message;
  message = "";
  m_tempo = ( ( uint32_t ) cpdata[2] ) << 16 |
            ( ( uint32_t ) cpdata[3] ) << 8  |
            ( ( uint32_t ) cpdata[4] );
    message += "Demain:";
  switch(m_tempo & 0x00490800){ //  Alerte : Bleu bit 18, Blanc bit 21, Rouge bit 24 info à 20H00
    case 0x00400000:
      message += "Bleu-";
    break;
    case 0x00080000:
      message += "Blanc-";
    break;
    case 0x00010000:
      message += "Rouge-";
    break;
    case 0x00000800:
      message += "Pas annonce-";
    break;
    default:
      message += "None-";
    break;
  }
  message += "Jour:";
  switch(cpdata[4] & 0x92){ // Jour : Bleu bit 33, Blanc bit 36, Rouge bit 39
  case 0x80:           // Jour : Bleu bit 33
    message += "Bleu-";
    break;
  case 0x10:           // Jour : Blanc bit 36
    message += "Blanc-";
    break;
  case 0x02:           // Jour : Rouge bit 39
    message += "Rouge-";
    break;
  default:
      message += "None-";
  break;
  }
  if ((m_tempo & 0x00024000)== 0x00020000)  // Heure Creuse Rouge bit 23 de 22H00 à 06H00
  {
    message += "HcR-";
  }
  if ((m_tempo & 0x00024000)== 0x00004000)  // Heure Pleine Rouge bit 26
  {
    message += "HpR-";    
  }
  if ((m_tempo & 0x00000120)== 0x00000100)  // Tarif Nuit tempo bit 32 de 22H00 à 06H00
  {
    message += "TN";
    code_acde |= _BV(Bit_Tn);
  }
  if ((m_tempo & 0x00000120)== 0x00000020)  // Tarif Jour tempo bit 35
  {
    message += "TJ";
    code_acde &= ~_BV(Bit_Tn);
  }

  Serial.println(message);
}

// Decodage plage Heures creuses A, C, D, E et detection Eclairage Public
// A : bits 1 & 2, C bits 3 & 4, D bits 8 & 9, E bits 11 & 12
// Eclairage bits 7, 10, 13, 16 
//
void set_rjn(void)
{
  //Serial.println("Set rjn");
  uint16_t m_acde;
  String message;
  code_acde = 0;
  M_data[0] = cpdata[0];
  M_data[1] = cpdata[1];
  M_data[2] = cpdata[2];
  M_data[3] = cpdata[3];
  M_data[4] = cpdata[4];
  M_data[5] = cpdata[5];  
 
  message = "";
  if ((M_data[0] & 0xC0)== 0x80) // HC A = 21H55 à 05H55
  {
    message += "A-";
    code_acde |= _BV(Bit_A);
  }
  else {
    message += "a-";
    code_acde &= ~_BV(Bit_A);
  }
  if ((M_data[0] & 0x30)== 0x20) // HC C = 21H25 à 05H25
  {
    message += "C-";
    code_acde |= _BV(Bit_C);
  }
  else {
    message += "c-";
    code_acde &= ~_BV(Bit_C);
  } 
  if ((M_data[0] & 0x01)== 0x01) // HC D = 22H55 à 06H55
  {
    message += "D-";
    code_acde |= _BV(Bit_D);
  }
  else {
    message += "d-";
    code_acde &= ~_BV(Bit_D);
  }
  if ((M_data[1] & 0x30)== 0x20) // HC E = 00H55 à 06H55 et 12H25 à 14H25
  {
    message += "E-";
    code_acde |= _BV(Bit_E);
  }
  else {
    message += "e-";
    code_acde &= ~_BV(Bit_E);
  } 
  code_acde &= 0x1F;
  if ((M_data[0] & 0x02)==0x02)
  {
    message += "00:00 => 04:00-";
  }
  switch (M_data[1] & 0x49){
  case 0x40:
    message += "Debut => Fin du Jour-";
    code_acde |= _BV(Bit_Jour);
    break;
  case 0x08:
    message += "Debut Nuit => 00:00-";
    code_acde |= _BV(Bit_Dn);
    break;
  case 0x01:
    message += "04:00 => Fin Nuit-";
    code_acde |= _BV(Bit_Fn);
    break;
  }
  Serial.print(message);
  set_tempo(); 
}
 
void decode_175(void)
{
    if ( det_175 < 0 ){
      uint8_t busy = 0;
      busy += parse_msg();
      //continue;
    }
    noInterrupts();
    ret1 = det_175;
    // m_det_175 = det_175;
    det_175 = -1;
    interrupts();
    ret = 0;
    if (ret1 > seuil ){
      ret++;
    }
    switch (state) {
    case D_REPOS:
      count = 0 ;
      nbset = 0 ;
      if ( ret ) {
        count++;
        state = D_START;
        ref_time = 0;
        next_step = ref_time +  CST_100;
      }
      break;
    case D_START:
      if ( ref_time < next_step ) {
        count += ret;
        break;
      }
      if ( count > 2 ) {
        state = D_FIRST;
        msg_bit_num = 0;
        start_time = ref_time +  CST_260;
        next_step = start_time +  CST_100;
        count = 0;
      } 
      else {
        state = D_REPOS;
      }
      break;
    case D_FIRST:
      if ( ref_time < start_time ) {
        break;
      }
      count += ret;
      if ( ref_time >= next_step ) {
        if ( count > 2 ) {
          nbset++;
          tempo_bit_store(msg_bit_num);
        }
        count = 0;
        start_time = next_step +  CST_150;
        next_step = start_time +  CST_100 ;
        if ( ++msg_bit_num >= NTBITS ) {
          state = D_REPOS;
          start_flag = 0;         
          tempo_clear_rdata();
          read_last_seq();
          set_rjn();
        }
        /* reset if no bit set after 4 bits */
        if ( nbset == 0 && msg_bit_num > 4 ) {
          state = D_REPOS;
        }
      }
      break;
    }
  }

void loop() {

}
